@@ -23,20 +23,35 @@ from telegrambotclient.handler import (
 from telegrambotclient.utils import pretty_format
 
 
-class ListRoute:
-    __slots__ = ("_route", )
+class DefaultRoute:
+    __slots__ = ("_root_route", "_name")
 
     def __init__(self, name: str, root_route):
-        if name not in root_route:
-            root_route[name] = []
-        self._route = root_route[name]
+        self._name = name
+        self._root_route = root_route
 
     def add_handler(self, handler: UpdateHandler):
-        self.add_handler_to_list(self._route, handler)
+        if self._name not in self._root_route:
+            self._root_route[self._name] = []
+        route = self._root_route[self._name]
+        for idx, _handler in enumerate(route):
+            if _handler.name == handler.name:
+                route[idx] = handler
+                return
+        route.append(handler)
+
+    def remove_handler(self, handler: UpdateHandler):
+        if self._name in self._root_route:
+            route = self._root_route[self._name]
+            for _handler in route:
+                if _handler.name == handler.name:
+                    route.remove(_handler)
+            if not route:
+                del self._root_route[self._name]
 
     async def call_handlers(self, bot: TelegramBot,
                             data: TelegramObject) -> bool:
-        for handler in self._route:
+        for handler in self._root_route.get(self._name, ()):
             if await self.call_handler(handler, bot, data) is bot.stop_call:
                 return bot.stop_call
         return bot.next_call
@@ -45,53 +60,64 @@ class ListRoute:
     async def call_handler(handler: UpdateHandler, *args, **kwargs) -> bool:
         return bool(await handler(*args, **kwargs))
 
-    @staticmethod
-    def add_handler_to_list(route, handler):
-        for idx, _handler in enumerate(route):
-            if _handler.name == handler.name:
-                route[idx] = handler
-                return route
-        route.append(handler)
 
-
-class ErrorRoute(ListRoute):
+class ErrorRoute(DefaultRoute):
     def __init__(self, root_route):
         super().__init__("error", root_route)
 
     async def call_handlers(self, bot: TelegramBot, data: TelegramObject,
                             error: Exception) -> bool:
-        for handler in self._route:
+        for handler in self._root_route.get("error", ()):
             if await self.call_handler(handler, bot, data,
                                        error) is bot.stop_call:
                 return bot.stop_call
         return bot.next_call
 
 
-class DictRoute(ListRoute):
-    def __init__(self, name: str, root_route):
-        if name not in root_route:
-            root_route[name] = {}
-        self._route = root_route[name]
-
-
-class CommandRoute(DictRoute):
+class CommandRoute(DefaultRoute):
     def __init__(self, root_route):
-        super().__init__("command", root_route)
+        super().__init__("cmd", root_route)
 
     def add_handler(self, handler: CommandHandler):
+        if self._name not in self._root_route:
+            self._root_route[self._name] = {}
         for cmd_word in handler.cmds:
-            self._route[cmd_word] = handler
+            self._root_route[self._name][cmd_word] = handler
+        route = self._root_route.get(UpdateField.MESSAGE.value, {})
+        route[self._name] = True
+        self._root_route[UpdateField.MESSAGE.value] = route
+        route = self._root_route.get(UpdateField.EDITED_MESSAGE.value, {})
+        route[self._name] = True
+        self._root_route[UpdateField.EDITED_MESSAGE.value] = route
+
+    def remove_handler(self, handler: CommandHandler):
+        if self._name in self._root_route:
+            route = self._root_route[self._name]
+            for cmd_word in handler.cmds:
+                if cmd_word in route:
+                    del route[cmd_word]
+            if not route:
+                del self._root_route[self._name]
+                if self._name in self._root_route.get(
+                        UpdateField.MESSAGE.value, {}):
+                    del self._root_route[UpdateField.MESSAGE.value][self._name]
+                if self._name in self._root_route.get(
+                        UpdateField.EDITED_MESSAGE.value, {}):
+                    del self._root_route[UpdateField.EDITED_MESSAGE.value][
+                        self._name]
 
     async def call_handlers(self, bot: TelegramBot, message: Message) -> bool:
         text = message.text
-        if not text or text[0] != '/' or not self._route:
+        if not text or text[0] != '/' or self._name not in self._root_route:
             return bot.next_call
         cmd_and_args = text.split()
         cmd_name_and_bot_username = cmd_and_args[0].split("@")
         if (len(cmd_name_and_bot_username) == 2
                 and cmd_name_and_bot_username[1] != bot.user.username):
             return bot.stop_call
-        handler = self._route.get(cmd_name_and_bot_username[0], None)
+        handler = self._root_route.get(self._name,
+                                       {}).get(cmd_name_and_bot_username[0],
+                                               None)
         if handler is None:
             return bot.next_call
         if len(cmd_and_args) == 1:
@@ -100,12 +126,22 @@ class CommandRoute(DictRoute):
                                        *cmd_and_args[1:])
 
 
-class ForceReplyRoute(DictRoute):
+class ForceReplyRoute(DefaultRoute):
     def __init__(self, root_route):
         super().__init__("force_reply", root_route)
 
     def add_handler(self, handler: ForceReplyHandler):
-        self._route[handler.name] = handler
+        if self._name not in self._root_route:
+            self._root_route[self._name] = {}
+        self._root_route[self._name][handler.name] = handler
+
+    def remove_handler(self, handler: ForceReplyHandler):
+        if self._name in self._root_route:
+            route = self._root_route[self._name]
+            if handler.name in route:
+                del route[handler.name]
+            if not route:
+                del self._root_route[self._name]
 
     async def call_handlers(self, bot: TelegramBot, message: Message) -> bool:
         force_reply_callback_name, force_reply_args = bot.get_force_reply(
@@ -113,10 +149,11 @@ class ForceReplyRoute(DictRoute):
             id if message.from_user else 0)
         if force_reply_callback_name is None:
             return bot.next_call
-        handler = self._route.get(force_reply_callback_name, None)
+        handler = self._root_route.get(self._name,
+                                       {}).get(force_reply_callback_name, None)
         if handler is None:
             raise TelegramBotException(
-                "{0} is not a force reply callback".format(
+                "{0} is not found as a force reply callback".format(
                     force_reply_callback_name))
         return await self.call_handler(
             handler, bot, message, *
@@ -124,115 +161,151 @@ class ForceReplyRoute(DictRoute):
                 handler, bot, message)
 
     def __contains__(self, force_reply_name: str):
-        return force_reply_name in self._route
+        return force_reply_name in self._root_route.get(self._name, {})
 
 
-class MessageRoute(DictRoute):
+class MessageRoute(DefaultRoute):
     def __init__(self, name: str, root_route):
         super().__init__(name, root_route)
 
     def add_handler(self, handler: _MessageHandler):
+        if self._name not in self._root_route:
+            self._root_route[self._name] = {}
+        route = self._root_route[self._name]
         message_fields = handler.message_fields
         if message_fields is None:
-            if "all" not in self._route:
-                self._route["all"] = []
-            self.add_handler_to_list(self._route["all"], handler)
+            DefaultRoute("all", route).add_handler(handler)
         elif isinstance(message_fields, str):
-            if "|" not in self._route:
-                self._route["|"] = defaultdict(list)
-            self.add_handler_to_list(self._route["|"][message_fields], handler)
+            if "|" not in route:
+                route["|"] = defaultdict(list)
+            DefaultRoute(message_fields, route["|"]).add_handler(handler)
         elif isinstance(message_fields, tuple):
-            if "|" not in self._route:
-                self._route["|"] = defaultdict(list)
+            if "|" not in route:
+                route["|"] = defaultdict(list)
             for field in handler.message_fields or ():
-                self.add_handler_to_list(self._route["|"][field], handler)
+                DefaultRoute(field, route["|"]).add_handler(handler)
         elif isinstance(message_fields, set):
-            if "&" not in self._route:
-                self._route["&"] = []
-            self.add_handler_to_list(self._route["&"], handler)
+            DefaultRoute("&", route).add_handler(handler)
+
+    def remove_handler(self, handler: _MessageHandler):
+        if self._name in self._root_route:
+            route = self._root_route[self._name]
+            message_fields = handler.message_fields
+            if message_fields is None:
+                DefaultRoute("all", route).remove_handler(handler)
+            elif isinstance(message_fields, str) and "|" in route:
+                DefaultRoute(message_fields,
+                             route["|"]).remove_handler(handler)
+            elif isinstance(message_fields, tuple) and "|" in route:
+                for field in handler.message_fields or ():
+                    DefaultRoute(field, route["|"]).remove_handler(handler)
+            elif isinstance(message_fields, set):
+                DefaultRoute("&", route).remove_handler(handler)
+            if not route:
+                del self._root_route[self._name]
 
     async def call_handlers(self, bot: TelegramBot, message: Message) -> bool:
-        # call 'and' handlers
-        handlers = self._route.get("&", None)
-        if handlers:
-            message_fields_set = set(message.keys())
-            for handler in handlers:
-                if handler.fields <= message_fields_set:
-                    if await self.call_handler(handler, bot,
-                                               message) is bot.stop_call:
-                        return bot.stop_call
-        # call 'or' handlers
-        handler_dict = self._route.get("|", None)
-        if handler_dict:
-            for message_field in handler_dict.keys():
-                if message_field in message:
-                    for handler in handler_dict[message_field]:
+        if self._name in self._root_route:
+            route = self._root_route[self._name]
+            # call 'and' handlers
+            handlers = route.get("&", None)
+            if handlers:
+                message_fields_set = set(message.keys())
+                for handler in handlers:
+                    if handler.fields <= message_fields_set:
                         if await self.call_handler(handler, bot,
                                                    message) is bot.stop_call:
                             return bot.stop_call
-        # call 'all' handlers
-        handlers = self._route.get("all", None)
-        if handlers:
-            for handler in handlers:
-                if await self.call_handler(handler, bot,
-                                           message) is bot.stop_call:
-                    return bot.stop_call
+            # call 'or' handlers
+            handler_dict = route.get("|", None)
+            if handler_dict:
+                for message_field in handler_dict.keys():
+                    if message_field in message:
+                        for handler in handler_dict[message_field]:
+                            if await self.call_handler(
+                                    handler, bot, message) is bot.stop_call:
+                                return bot.stop_call
+            # call 'all' handlers
+            handlers = route.get("all", None)
+            if handlers:
+                for handler in handlers:
+                    if await self.call_handler(handler, bot,
+                                               message) is bot.stop_call:
+                        return bot.stop_call
         return bot.next_call
 
 
-class CallbackQueryRoute(DictRoute):
+class CallbackQueryRoute(DefaultRoute):
     def __init__(self, root_route):
         super().__init__(UpdateField.CALLBACK_QUERY.value, root_route)
 
     def add_handler(self, handler: CallbackQueryHandler):
+        if self._name not in self._root_route:
+            self._root_route[self._name] = {}
+        route = self._root_route[self._name]
         has_callback_data, has_callback_data_name, has_callback_data_regex, has_callback_data_parse = handler.have_matchers
         if has_callback_data:
-            if "data" not in self._route:
-                self._route["data"] = {}
-            self._route["data"][handler.callback_data] = handler
+            if "data" not in route:
+                route["data"] = {}
+            route["data"][handler.callback_data] = handler
         if has_callback_data_name:
-            if "name" not in self._route:
-                self._route["name"] = {}
-            self._route["name"][handler.callback_data_name] = handler
+            if "name" not in route:
+                route["name"] = {}
+            route["name"][handler.callback_data_name] = handler
         if has_callback_data_regex:
-            if "regex" not in self._route:
-                self._route["regex"] = []
-            self.add_handler_to_list(self._route["regex"], handler)
+            DefaultRoute("regex", route).add_handler(handler)
         if has_callback_data_parse:
-            if "parse" not in self._route:
-                self._route["parse"] = []
-            self.add_handler_to_list(self._route["parse"], handler)
+            DefaultRoute("parse", route).add_handler(handler)
+
+    def remove_handler(self, handler: CallbackQueryHandler):
+        if self._name in self._root_route:
+            route = self._root_route[self._name]
+            has_callback_data, has_callback_data_name, has_callback_data_regex, has_callback_data_parse = handler.have_matchers
+            if has_callback_data and "data" in route and handler.callback_data in route[
+                    "data"]:
+                del route["data"][handler.callback_data]
+            if has_callback_data_name and "name" in route and handler.callback_data_name in route[
+                    "name"]:
+                del route["name"][handler.callback_data_name]
+            if has_callback_data_regex:
+                DefaultRoute("regex", route).remove_handler(handler)
+            if has_callback_data_parse:
+                DefaultRoute("parse", route).remove_handler(handler)
+            if not route:
+                del self._root_route[self._name]
 
     async def call_handlers(self, bot: TelegramBot,
                             callback_query: CallbackQuery) -> bool:
         if callback_query.data is None:
             raise TelegramBotException("callback_query.data is None")
-        # for callback_data
-        handler = self._route.get("data", {}).get(callback_query.data, None)
-        if handler and await self.call_handler(
-                handler, bot, callback_query) is bot.stop_call:
-            return bot.stop_call
-        #for callback_data_name
-        callback_name_data = callback_query.data.split("|")
-        if len(callback_name_data) == 2:
-            callback_name, callback_data = tuple(callback_name_data)
-            handler = self._route.get("name", {}).get(callback_name, None)
+        if self._name in self._root_route:
+            route = self._root_route[self._name]
+            # for callback_data
+            handler = route.get("data", {}).get(callback_query.data, None)
             if handler and await self.call_handler(
-                    handler, bot, callback_query, *
-                    json.loads(callback_data)) is bot.stop_call:
-                return bot.stop_call
-        # for callback_data_regex
-        for handler in self._route.get("regex", ()):
-            result = handler.callback_data_match(callback_query)
-            if result and await handler(bot, callback_query,
-                                        result) is bot.stop_call:
-                return bot.stop_call
-        # for callback_data_parse
-        for handler in self._route.get("parse", ()):
-            result = handler.callback_data_parse(callback_query)
-            if result and await self.call_handler(handler, bot, callback_query,
-                                                  result) is bot.stop_call:
+                    handler, bot, callback_query) is bot.stop_call:
                 return bot.stop_call
+            #for callback_data_name
+            callback_name_data = callback_query.data.split("|")
+            if len(callback_name_data) == 2:
+                callback_name, callback_data = tuple(callback_name_data)
+                handler = route.get("name", {}).get(callback_name, None)
+                if handler and await self.call_handler(
+                        handler, bot, callback_query, *
+                        json.loads(callback_data)) is bot.stop_call:
+                    return bot.stop_call
+            # for callback_data_regex
+            for handler in route.get("regex", ()):
+                result = handler.callback_data_match(callback_query)
+                if result and await self.call_handler(
+                        handler, bot, callback_query, result) is bot.stop_call:
+                    return bot.stop_call
+            # for callback_data_parse
+            for handler in route.get("parse", ()):
+                result = handler.callback_data_parse(callback_query)
+                if result and await self.call_handler(
+                        handler, bot, callback_query, result) is bot.stop_call:
+                    return bot.stop_call
         return bot.next_call
 
 
@@ -298,7 +371,36 @@ class TelegramRouter:
         if isinstance(handler, ErrorHandler):
             return ErrorRoute(self._route_map).add_handler(handler)
         # for others update handlers
-        return ListRoute(update_field, self._route_map).add_handler(handler)
+        return DefaultRoute(update_field, self._route_map).add_handler(handler)
+
+    def remove_handler(self, handler: UpdateHandler):
+        assert isinstance(handler, UpdateHandler), True
+        update_field = handler.update_field
+        logger.info("unbind a %s handler: '%s@%s'", update_field, handler.name,
+                    self.name)
+        if isinstance(handler, CommandHandler):
+            return CommandRoute(self._route_map).remove_handler(handler)
+        if isinstance(handler, ForceReplyHandler):
+            return ForceReplyRoute(self._route_map).remove_handler(handler)
+        if isinstance(handler, MessageHandler):
+            return MessageRoute(UpdateField.MESSAGE.value,
+                                self._route_map).remove_handler(handler)
+        if isinstance(handler, EditedMessageHandler):
+            return MessageRoute(UpdateField.EDITED_MESSAGE.value,
+                                self._route_map).remove_handler(handler)
+        if isinstance(handler, CallbackQueryHandler):
+            return CallbackQueryRoute(self._route_map).remove_handler(handler)
+        if isinstance(handler, ChannelPostHandler):
+            return MessageRoute(UpdateField.CHANNEL_POST.value,
+                                self._route_map).remove_handler(handler)
+        if isinstance(handler, EditedChannelPostHandler):
+            return MessageRoute(UpdateField.EDITED_CHANNEL_POST.value,
+                                self._route_map).add_handler(handler)
+        if isinstance(handler, ErrorHandler):
+            return ErrorRoute(self._route_map).remove_handler(handler)
+        # for others update handlers
+        return DefaultRoute(update_field,
+                            self._route_map).remove_handler(handler)
 
     def register_error_handler(self, callback: Callable, errors: Union[List,
                                                                        Tuple]):
@@ -533,7 +635,7 @@ class TelegramRouter:
 
     async def call_message_handlers(self, bot: TelegramBot,
                                     message: Message) -> bool:
-        if await CommandRoute(self._route_map).call_handlers(
+        if message.text and await CommandRoute(self._route_map).call_handlers(
                 bot, message) is self.stop_call:
             return self.stop_call
         if await ForceReplyRoute(self._route_map).call_handlers(
@@ -545,14 +647,14 @@ class TelegramRouter:
 
     async def call_edited_message_handlers(self, bot: TelegramBot,
                                            edited_message: Message) -> bool:
-        if await CommandRoute(self._route_map).call_handlers(
-                bot, edited_message) is self.stop_call:
-            return self.stop_call
-        # for a edited message, the fields should be text ONLY for force_reply
-        if edited_message.text and await ForceReplyRoute(
-                self._route_map).call_handlers(
+        if edited_message.text:
+            if await CommandRoute(self._route_map).call_handlers(
                     bot, edited_message) is self.stop_call:
-            return self.stop_call
+                return self.stop_call
+            # for a edited message, the fields should be text ONLY for force_reply
+            if await ForceReplyRoute(self._route_map).call_handlers(
+                    bot, edited_message) is self.stop_call:
+                return self.stop_call
         return await MessageRoute(UpdateField.EDITED_MESSAGE.value,
                                   self._route_map).call_handlers(
                                       bot, edited_message)
@@ -574,46 +676,46 @@ class TelegramRouter:
 
     async def call_inline_query_handlers(self, bot: TelegramBot,
                                          inline_query: InlineQuery) -> bool:
-        return await ListRoute(UpdateField.INLINE_QUERY.value,
-                               self._route_map).call_handlers(
-                                   bot, inline_query)
+        return await DefaultRoute(UpdateField.INLINE_QUERY.value,
+                                  self._route_map).call_handlers(
+                                      bot, inline_query)
 
     async def call_chosen_inline_result_handlers(
             self, bot: TelegramBot,
             chosen_inline_result: ChosenInlineResult) -> bool:
-        return await ListRoute(UpdateField.CHOSEN_INLINE_RESULT.value,
-                               self._route_map).call_handlers(
-                                   bot, chosen_inline_result)
+        return await DefaultRoute(UpdateField.CHOSEN_INLINE_RESULT.value,
+                                  self._route_map).call_handlers(
+                                      bot, chosen_inline_result)
 
     async def call_shipping_query_handlers(
             self, bot: TelegramBot, shipping_query: ShippingQuery) -> bool:
-        return await ListRoute(UpdateField.SHIPPING_QUERY.value,
-                               self._route_map).call_handlers(
-                                   bot, shipping_query)
+        return await DefaultRoute(UpdateField.SHIPPING_QUERY.value,
+                                  self._route_map).call_handlers(
+                                      bot, shipping_query)
 
     async def call_pre_checkout_query_handlers(
             self, bot: TelegramBot,
             pre_checkout_query: PreCheckoutQuery) -> bool:
-        return await ListRoute(UpdateField.PRE_CHECKOUT_QUERY.value,
-                               self._route_map).call_handlers(
-                                   bot, pre_checkout_query)
+        return await DefaultRoute(UpdateField.PRE_CHECKOUT_QUERY.value,
+                                  self._route_map).call_handlers(
+                                      bot, pre_checkout_query)
 
     async def call_poll_handlers(self, bot: TelegramBot, poll: Poll) -> bool:
-        return await ListRoute(UpdateField.POLL.value,
-                               self._route_map).call_handlers(bot, poll)
+        return await DefaultRoute(UpdateField.POLL.value,
+                                  self._route_map).call_handlers(bot, poll)
 
     async def call_poll_answer_handlers(self, bot: TelegramBot,
                                         poll_answer: PollAnswer) -> bool:
-        return await ListRoute(UpdateField.POLL_ANSWER.value,
-                               self._route_map).call_handlers(
-                                   bot, poll_answer)
+        return await DefaultRoute(UpdateField.POLL_ANSWER.value,
+                                  self._route_map).call_handlers(
+                                      bot, poll_answer)
 
     async def call_my_chat_member_handlers(
             self, bot: TelegramBot,
             my_chat_member_updated: ChatMemberUpdated) -> bool:
-        return await ListRoute(UpdateField.MY_CHAT_MEMBER,
-                               self._route_map).call_handlers(
-                                   bot, my_chat_member_updated)
+        return await DefaultRoute(UpdateField.MY_CHAT_MEMBER,
+                                  self._route_map).call_handlers(
+                                      bot, my_chat_member_updated)
 
     async def call_chat_member_handlers(
             self, bot: TelegramBot,
